use config::{Config, File};
use glob::glob;
use lazy_static::lazy_static;
use log::LevelFilter;
use poise::serenity_prelude as serenity;
use poise::serenity_prelude::UserId;
use regex::Regex;
use simplelog::SimpleLogger;
use std::{collections::HashSet, sync::RwLock};

// mod entity;

// Types used by all command functions
type Error = Box<dyn std::error::Error + Send + Sync>;
type Context<'a> = poise::Context<'a, Data, Error>;

mod commands;

// Custom user data passed to all command functions
pub struct Data {
    database: sqlx::SqlitePool,
}

lazy_static! {
    static ref SETTINGS: RwLock<Config> = RwLock::new(Config::default());
}

/// Register application commands in this guild or globally
///
/// Run with no arguments to register in guild, run with argument "global" to register globally.
#[poise::command(
    prefix_command,
    hide_in_help,
    // required_permissions = "MANAGE_MESSAGES | ADMINISTRATOR",
    owners_only=true,
)]
async fn register(ctx: Context<'_>, #[flag] global: bool) -> Result<(), Error> {
    poise::builtins::register_application_commands(ctx, global).await?;

    Ok(())
}

/// Boop the bot!
#[poise::command(prefix_command, track_edits, slash_command)]
pub async fn boop(ctx: Context<'_>) -> Result<(), Error> {
    let uuid_boop = ctx.id();

    ctx.send(|m| {
        m.content("I want some boops!").components(|c| {
            c.create_action_row(|ar| {
                ar.create_button(|b| {
                    b.style(serenity::ButtonStyle::Primary)
                        .label("Boop me!")
                        .custom_id(uuid_boop)
                })
            })
        })
    })
    .await?;

    let mut boop_count = 0;
    while let Some(mci) = serenity::CollectComponentInteraction::new(ctx.discord())
        .author_id(ctx.author().id)
        .channel_id(ctx.channel_id())
        .timeout(std::time::Duration::from_secs(1200))
        .filter(move |mci| mci.data.custom_id == uuid_boop.to_string())
        .await
    {
        boop_count += 1;

        let mut msg = mci.message.clone();
        msg.edit(ctx.discord(), |m| {
            m.content(format!("Boop count: {}", boop_count))
        })
        .await?;

        mci.create_interaction_response(ctx.discord(), |ir| {
            ir.kind(serenity::InteractionResponseType::DeferredUpdateMessage)
        })
        .await?;
    }

    Ok(())
}

async fn on_error(error: poise::FrameworkError<'_, Data, Error>) {
    // This is our custom error handler
    // They are many errors that can occur, so we only handle the ones we want to customize
    // and forward the rest to the default handler
    match error {
        poise::FrameworkError::Setup { error } => panic!("Failed to start bot: {:?}", error),
        poise::FrameworkError::Command { error, ctx } => {
            println!("Error in command `{}`: {:?}", ctx.command().name, error,);
        }
        error => {
            if let Err(e) = poise::builtins::on_error(error).await {
                println!("Error while handling error: {}", e)
            }
        }
    }
}

/// Show this help menu
#[poise::command(prefix_command, track_edits, slash_command)]
async fn help(
    ctx: Context<'_>,
    #[description = "Specific command to show help about"]
    #[autocomplete = "poise::builtins::autocomplete_command"]
    command: Option<String>,
) -> Result<(), Error> {
    poise::builtins::help(
        ctx,
        command.as_deref(),
        poise::builtins::HelpConfiguration {
            extra_text_at_bottom: "\
This bot is very much a work in progress. Commands will change and may break.",
            show_context_menu_commands: true,
            ..Default::default()
        },
    )
    .await?;
    Ok(())
}

fn setup_config() -> Result<(), Error> {
    SETTINGS
        .write()
        .unwrap()
        .merge(
            glob("data/config/*")
                .unwrap()
                .map(|path| File::from(path.unwrap()))
                .collect::<Vec<_>>(),
        )
        .unwrap();

    println!("{:#?}", SETTINGS.read());
    Ok(())
}

async fn app() -> Result<(), Error> {
    let discord = SETTINGS.read().unwrap().get_table("discord").unwrap();

    let mut owners: HashSet<UserId> = HashSet::new();
    owners.insert(UserId(117992484310745097));

    let prefix: String = discord
        .get("prefix")
        .expect("Config error, please set the discord[token] value")
        .clone()
        .into_str()
        .expect("Config error, please make sure discord[token] is a string");
    let options = poise::FrameworkOptions {
        commands: vec![
            help(),
            register(),
            commands::invites::invites(),
            commands::schedule::schedule(),
            commands::boop(),
            commands::planning::get_command(),
            // poise::Command {
            //     subcommands: vec![
            //         commands::planning::series(),
            //         commands::planning::sessions(),
            //         commands::planning::events(),
            //         // Let's make sure poise isn't confused by the duplicate names!
            //         // commands::planning::planning(),
            //     ],
            //     ..commands::planning::planning()
            // },
        ],
        prefix_options: poise::PrefixFrameworkOptions {
            prefix: Some(prefix.into()),
            edit_tracker: Some(poise::EditTracker::for_timespan(
                std::time::Duration::from_secs(3600),
            )),
            additional_prefixes: vec![
                poise::Prefix::Regex(Regex::new("(?i)ol'? rusty please").unwrap()),
                poise::Prefix::Regex(Regex::new("(?i)Oi Rusty please").unwrap()),
            ],
            ..Default::default()
        },
        on_error: |error| Box::pin(on_error(error)),
        pre_command: |ctx| {
            Box::pin(async move {
                println!("Executing command {}...", ctx.command().qualified_name);
            })
        },
        post_command: |ctx| {
            Box::pin(async move {
                println!("Executed command {}!", ctx.command().qualified_name);
            })
        },
        owners,
        ..Default::default()
    };

    // let discord = SETTINGS.read().unwrap().get_table("discord").unwrap();
    let token = discord
        .get("token")
        .expect("Config error, please set the discord[token] value")
        .clone()
        .into_str()
        .expect("Config error, please make sure discord[token] is a string");

    let database = sqlx::sqlite::SqlitePoolOptions::new()
        .max_connections(5)
        .connect_with(
            "sqlite:data/database.db"
                .parse::<sqlx::sqlite::SqliteConnectOptions>()?
                .create_if_missing(true),
        )
        .await?;
    sqlx::migrate!("./migrations").run(&database).await?;

    poise::Framework::build()
        .token(token)
        .user_data_setup(move |_ctx, _ready, _framework| {
            Box::pin(async move { Ok(Data { database: database }) })
        })
        .options(options)
        .run()
        .await
        .unwrap();

    Ok(())
}

#[tokio::main]
async fn main() {
    print!("Starting the bot");
    setup_config().unwrap();
    // env_logger::init();
    // init_log();
    SimpleLogger::init(LevelFilter::Warn, simplelog::Config::default()).unwrap();

    if let Err(e) = app().await {
        log::error!("{}", e);
        std::process::exit(1);
    }
}
